/*
  html.cc

  Copyright (c) 1996 Roland Wunderling, Malte Zoeckler
  Copyright (c) 1998 Michael Meeks
  Copyright (c) 1998-2000 Dragos Acostachioaie

  This file is part of DOC++.

  DOC++ is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the license, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <assert.h>
#include <ctype.h>
#ifdef __BORLANDC__
#include <dir.h>
#endif
#if defined(__VISUALC__) || defined(__WATCOMC__)
#include <direct.h>
#endif
#include <errno.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "McHashTable.h"
#include "McSorter.h"
#include "classgraph.h"
#include "doc.h"
#include "gifs.h"
#include "java.h"

extern bool	withTables;			// --tables
extern bool	withBorders;			// --tables-border
extern char*    ownFooter;			// --footer FILE
extern bool	javaGraphs;			// --no-java-graphs
extern bool	trivialGraphs;			// --trivial-graphs
extern bool	useGeneral;			// --no-general
extern bool	sortEntries;			// --sort
extern char*    ownHeader;			// --header FILE
extern char*    htmlSuffix;			// --suffix SUF (default: ".html")

#define max(a,b) ((a)>(b) ? (a) : (b))

#define BGCOLOR "<BODY BGCOLOR=\"#ffffff\">\n"

static McString header, footer;
static char *GENERAL_NAME = "General";
FILE *generalf;

static McString indexHeader("<HTML>\n<HEAD>\n   <TITLE>Table of Contents</TITLE>\n"
			   "   <META NAME=\"GENERATOR\" CONTENT=\"DOC++ " DOCXX_VERSION "\">\n"
			   "</HEAD>\n" BGCOLOR);
static McString indexFooter;

static McString hierHeader("<HTML>\n<HEAD>\n   <TITLE>Hierarchy of Classes</TITLE>\n"
			   "   <META NAME=\"GENERATOR\" CONTENT=\"DOC++ " DOCXX_VERSION "\">\n"
			   "</HEAD>\n" BGCOLOR);
static McString hierFooter;

static McString generalHeader("<HTML>\n<HEAD>\n   <TITLE>General Bits</TITLE>\n"
			   "   <META NAME=\"GENERATOR\" CONTENT=\"DOC++ " DOCXX_VERSION "\">\n"
			   "</HEAD>\n" BGCOLOR);

static McString pageHeader(BGCOLOR);
static McString pageFooter;
static McString pageFooterJava;

void writeTOCentry(FILE *f, Entry *e, int memo, int dup = 0);

int makedir(const char *d, int perm)
{
#if defined(__BORLANDC__) || defined(__WATCOMC__)
    return mkdir(d);
#else
#ifdef __VISUALC__
    return _mkdir(d);
#else
    return mkdir(d, perm);
#endif
#endif
}

#define myisalnum(c) ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || \
		      (c >= '0' && c <= '9') || c == '_' || c== '.' || c == '-')

// Initializes HTML footers
void buildFooters()
{
    indexFooter = "<I><A HREF=\"HIER";
    indexFooter += htmlSuffix;
    indexFooter += "\">Hierarchy of classes</A></I><P>";

    hierFooter = "<I><A HREF=\"index";
    hierFooter += htmlSuffix;
    hierFooter += "\">Alphabetic index</A></I><P>";

    pageFooter = "<I><A HREF=\"index";
    pageFooter += htmlSuffix;
    pageFooter += "\">Alphabetic index</A></I>";
    pageFooter += " <I><A HREF=\"HIER";
    pageFooter += htmlSuffix;
    pageFooter += "\">Hierarchy of classes</A></I></P>";

    pageFooterJava = "<I><A HREF=\"index";
    pageFooterJava += htmlSuffix;
    pageFooterJava += "\">Alphabetic index</A></I>";
    pageFooterJava += " <I><A HREF=\"HIER";
    pageFooterJava += htmlSuffix;
    pageFooterJava += "\">HTML hierarchy of classes</A>";
    pageFooterJava += " or <A HREF=\"HIERjava";
    pageFooterJava += htmlSuffix;
    pageFooterJava += "\">Java</A></I><P>";
}

McString makeFileName(const McString& str, Entry *e)
{
    static McHashTable<char *,int> files(1);
    McString s, ls;
    char buf[40];
    int i, l = str.length();
    int odd = 0;	// An unusual thing ...

    for(i = 0; i < l; i++)
	if(myisalnum(str[i]))
	    {
	    char c = str[i];
	    s += c;
	    if(c >= 'A' && c <= 'Z')
		c += (((int)'a') - 'A');
	    ls += c;
	    }
    char *tmp = strdup(ls.c_str());
    int *val = files.insert(tmp);
    if(*val > 1)
	{
	free(tmp);
	s += '.';
	sprintf(buf, "%d", *val);
	s += buf;
	odd = 1;
	}
    (*val)++;

    s += htmlSuffix;

#ifdef DEBUG
    if(verb && odd)
	{
	fprintf(stderr, "Warning: weird file name `%s'\n", (char *)s);
	e->dump(stdout);
	}
#endif
    return s;
}

enum _outType
{
    DOC		= 1,
    MEMO	= 2
};

void htmlComment(FILE *f, Entry *e, int t)
{
    bool done = false;
    char start[] = "<BLOCKQUOTE>";
    char end[] = "</BLOCKQUOTE>";

    if((t & DOC) && e->doc.length() > 1)
	{
	fprintf(f, "%s%s%s\n", start, (e->hdoc), end);
	done = true;
	}
    if((t & MEMO) && !done && e->memo.length() > 1)
	fprintf(f, "%s%s%s\n", start, (e->hmemo), end);
}

int subEntryIsToBeDocumented(Entry *entry)
{
    bool toBeDocumented = false;

    if(entry->protection != PRIV || withPrivate)
	{
	if(entry->retrn.size() ||
	    entry->param.size() || entry->author.length() ||
	    entry->see.size() || entry->version.length())
	    toBeDocumented = true;
	if(entry->doc.length() > entry->memo.length() || 
	    (entry->doc.length() == entry->memo.length() &&
	     entry->doc == entry->memo))
	    toBeDocumented = true;

	if(alwaysPrintDocSection)
	    toBeDocumented = true;
	}

    return toBeDocumented;
}

// Writes classgraphs. Either as text-graphs or as java-graphs.
class ClassGraphWriter
{
    public:
	static void writeJava(FILE *f, Entry *e, bool directOnly = true);
	static void writeText(FILE *f, Entry *e);
	static void write(FILE *f, Entry *e);
	static void writeImplements(FILE *f, Entry *e);
};

// This class writes members.
class MemberWriter
{
    public:
	MemberWriter()
	    {
	    f = NULL;
	    }
    protected:
	McString heading;
	int first;
	FILE *f;

	virtual const char *startString()
	    {
	    return "<P><DL>";
	    };
	virtual const char *endString()
	    {
	    return "</DL></P>";
	    };
	virtual void showSubMembers(Entry *);
	struct ListEntry {
	    Entry *entry;
	    int links;
	    int withSub;
	};
	McDArray<ListEntry> list;
	virtual void writeMember(Entry *e, bool links, bool withSub = true);
	class EntryCompare {
	    public:
		int operator()(const ListEntry& l1, const ListEntry& l2)
		    {
		    return strcmp(l1.entry->name.c_str(), l2.entry->name.c_str());
		    }
	};
    public:
	virtual void sort()
	    {
	    EntryCompare comp;
	    if(list.size())
		::sort((ListEntry *)list, list.size(), comp, 0);
	    }
	virtual void startList(FILE *f, char *heading, bool withLinks);
	virtual void addMember(Entry *e, bool links, bool withSub = true)
	    {
	    ListEntry le;
	    le.entry = e;
	    le.links = links;
	    le.withSub = withSub;
	    list.append(le);
	    }
	virtual void endList();
	virtual ~MemberWriter()
	    {};
};

class MemberWriterTable : public MemberWriter
{
    protected:
	virtual const char *startString()
	    {
	    return (withBorders ? "<P><TABLE BORDER>" : "<P><TABLE>");
	    };

	virtual const char *endString()
	    {
	    return "</TABLE></P>";
	    };

	virtual void writeMember(Entry *e, bool links, bool withSub = true);
};

// Renamed from `SortedList' as far too specific now...
class TOClist : public MemberWriter
{
    protected:
	struct ListEntry
	    {
	    Entry *e;
	    TOClist *tl;
	    const char *name;
	    };
	McDArray<ListEntry> list;
    public:
	virtual void sort()
	    {
	    EntryCompare comp;
	    int lp;

	    if(list.size() > 1)
		::sort((ListEntry *)list, list.size(), comp, 0);

	    // Sort subsections
	    for(lp = 0; lp < list.size(); lp++)
		if(list[lp].tl)
	    list[lp].tl->sort();
	    }
	class EntryCompare
	    {
		public:
		    int operator()(ListEntry& l1, ListEntry& l2)
			{
			return strcmp(l1.name, l2.name);
			}
	    };
	void addEntry(Entry *entry, TOClist *tl);
	void write(FILE *f);
	int size()
	    {
	    return list.size();
	    }
	~TOClist();
};

void TOClist::write(FILE *f)
{
    // Resolve conflicts, get meaningful data from WriteTOCentry
    //    (f, entry, 1), and go for it... merge writeTOCEntry into this ?
    // As a further burden, try and merge duplicate ClassName link references
    // ALSO, do sub lists
    int i, last_collision = 0, size = list.size(), conf;
    bool inUL = false;

    for(i = 0; i < size; i++)
	{
	ListEntry *le, *nle;
	le = &list[i];

	if(i < size - 1)
	    nle = &list[i + 1];
	else
	    nle = 0;

	// Skip identical entries (e.g. loads of constructors in a class)
	// ---
	// Global functions that are overloaded get seen as indentical,
	// but they in fact have different documentation, which is why we 
	// check for fileName as well <cpbotha@bigfoot.com>
	if(nle && strcmp(le->name, nle->name) == 0 && le->tl == nle->tl)
	    continue;

	// Detect identicaly named entries (e.g. lots of next())
	// ---
	// See comment above about overloaded global functions
	// <cpbotha@bigfoot.com>
	conf = nle && (strcmp(le->e->fullName.c_str(), nle->e->fullName.c_str()) == 0);
	if(conf != last_collision && conf)
	    {
	    fprintf(f, "<LI><B>%s:</B>\n<UL>\n", le->e->name.c_str());
	    inUL = true;
	    }

	fprintf(f, "<LI>");
	writeTOCentry(f, le->e, 1, (conf | last_collision));

	if(conf != last_collision && !conf)
	    {
	    fprintf(f, "</UL>\n");
	    inUL = false;
	    }

	if(le->tl)	// A dreaded sub-list e.g. package list
	    {
	    fprintf(f, "<UL>");
	    le->tl->write(f);
	    fprintf(f, "</UL>");
	    }
	last_collision = conf;
	}
    if(inUL)		// Some odd thing happened at the end of the list...
	fprintf(f, "</UL>\n");
}

void TOClist::addEntry(Entry *entry, TOClist *tl)
{
    ListEntry le;

    // don't show class members in the TOC... due to popular demand
    // do the same with structure and namespace members
    if(!showMembersInTOC)
	if((entry->parent->section & CLASS_SEC) ||
	    entry->parent->section == NAMESPACE_SEC ||
	    entry->parent->section == UNION_SEC ||
	    (entry->parent->section == MANUAL_SEC && entry->parent != root))
	    return;

    le.e = entry;
    le.tl = tl;
    le.name = entry->fullName.c_str();
    list.append(le);
}

TOClist::~TOClist()
{
    int i;

    for(i = 0; i < list.size(); i++)
	if(list[i].tl)
	    free(list[i].tl);
}

void MemberWriter::showSubMembers(Entry *e)
{
    Entry *tmp;
    fprintf(f, "%s\n", startString());
    for(tmp = e->sub; tmp; tmp = tmp->next)
	writeMember(tmp, 1);
    fprintf(f, "%s\n", endString());
}

void printRefLabel(FILE *f, Entry *entry)
{
    fprintf(f, "\n<A NAME=\"%s\">\n<A NAME=\"DOC.", entry->name.c_str());
    entry->dumpNumber(f);
    fprintf(f, "\">\n");
}

void MemberWriter::writeMember(Entry *entry, bool link, bool withSub)
{
    if(first)
	{
	fprintf(f, "%s\n%s", startString(), heading.c_str());
	first = 0;
	}

    if(!subEntryIsToBeDocumented(entry) && link)
	printRefLabel(f, entry);

    if((entry->section & CLASS_SEC) || entry->section == NAMESPACE_SEC)
	{
	fprintf(f, "<DT><A HREF=\"%s\">%s <B>%s</B></A>\n", entry->fileName.c_str(),
	    entry->htype, entry->hname);
	fprintf(f, "<DD><I>%s</I>\n", entry->hmemo);
	}
    else
	{
	fprintf(f, "<DT>\n");
	if(subEntryIsToBeDocumented(entry) && link)
	    {
	    fprintf(f, "<A HREF=\"#DOC.");
	    entry->dumpNumber(f);
	    fprintf(f, "\"><IMG ALT=\"[more]\" BORDER=0 SRC=icon1.gif></A>");
	    }
	else
	    if(entry->section != MANUAL_SEC)
		fprintf(f, "<IMG ALT=\"o\" SRC=icon2.gif>");
	    else 
        	fprintf(f, "<P>");

	fprintf(f, "%s ", entry->htype);
	fprintf(f, "<B>%s</B>%s\n", entry->hname, entry->hargs);

	if(link)
	    fprintf(f, "<DD><I>%s</I>\n", entry->hmemo);
	}

    if(entry->sub && withSub)
	showSubMembers(entry);
}

void MemberWriter::startList(FILE *file, char *head, bool withLinks)
{
    f = file;
    heading = head;
    first = 1;
    list.resize(0);
}

void MemberWriter::endList()
{
    int i;

    if(f == NULL)
	return;

    if(sortEntries)
	sort();

    for(i = 0; i < list.size(); i++)
	writeMember(list[i].entry, list[i].links, list[i].withSub);

    if(!first)
	fprintf(f, "%s\n\n", endString());
}

void MemberWriterTable::writeMember(Entry *entry, bool link, bool withSub)
{
    char *args, *type;

    if(first)
	{
	fprintf(f, "%s\n%s", startString(), heading.c_str());
	first = 0;
	}

    fprintf(f, "<TR>");
    fprintf(f, "<TD VALIGN=top>");

    if(!subEntryIsToBeDocumented(entry) && link) 
	printRefLabel(f, entry);

    if((entry->section & CLASS_SEC) || entry->section == NAMESPACE_SEC)
	{
	fprintf(f, "<DT><A HREF=\"%s\">%s <B>%s</B></A>\n", entry->fileName.c_str(),
	    entry->htype, entry->hname);
	fprintf(f, "<DD><I>%s</I>\n", entry->hmemo);
	}
    else
	{
	if(subEntryIsToBeDocumented(entry) && link)
	    {
	    fprintf(f, "<A HREF=\"#DOC.");
	    entry->dumpNumber(f);
	    fprintf(f, "\"><IMG ALT=\"[more]\" BORDER=0 SRC=icon1.gif></A>");
	    }
	else
	    if(entry->section != MANUAL_SEC)
		fprintf(f, "<IMG ALT=\"o\" SRC=icon2.gif>");
	    else
		fprintf(f, "<P>");

	args = entry->hargs;
	type = entry->htype;

	fprintf(f, "%s ", type);
	fprintf(f, "</TD><TD>");
	fprintf(f, "<B>%s</B> %s<BR>", entry->hname, args);

	if(link)
	    fprintf(f, "\n<I>%s</I>\n", entry->hmemo);
	}

    if(entry->sub && withSub)
	showSubMembers(entry);

    fprintf(f, "</TD></TR>");
}

void ClassGraphWriter::writeJava(FILE *f, Entry *entry, bool directOnly)
{
    ClassGraph cg(entry, 0);
    ClassGraph *cls = &cg;
    McString classes, before, after, indent;
    char first = 1;
    int numLines = 0, longest = 0;
    char buf[40];

    cg.addBases();

    if(directOnly)
	cg.addDirectChilds();
    else
	cg.addAllChilds();

    for(cls = cg.firstLine; cls; cls = cls->nextLine)
	{
	numLines++;    
	if(first)
	    first = 0;
	else
	    {
	    classes += ",";
	    before += ",";
	    after += ",";
	    indent += ",";
	    }
	if(cls->entry)
	    {
	    if(longest < cls->entry->name.size())
		longest = cls->entry->name.size();

	    if(cls->entry->section & CLASS_SEC)
		classes += "C";
	    else
		classes += "I";
	    classes += cls->entry->name;
	    classes += ",M";
	    classes += cls->entry->fileName;
	    }
	else
	    {
	    if(longest < cls->name.size())
		longest = cls->name.size();
	    classes += "M" + cls->name;
	    classes += ",M";
	    }
	before += "M" + cls->before;
	after += "M" + cls->after;
	sprintf(buf, "%d", cls->indent);
	indent += buf;
	}

    fprintf(f, "<APPLET CODE=\"ClassGraph.class\" WIDTH=600 HEIGHT=%d>\n",
	numLines * 30 + 5);
    fprintf(f, "<param name=classes value=\"%s\">\n", classes.c_str());
    fprintf(f, "<param name=before value=\"%s\">\n", before.c_str());
    fprintf(f, "<param name=after value=\"%s\">\n", after.c_str());
    fprintf(f, "<param name=indent value=\"%s\">\n", indent.c_str());
    fprintf(f, "<param name=arrowdir value=");
    if(upArrows)
	fprintf(f, "\"up\">\n");
    else
	fprintf(f, "\"down\">\n");

    fprintf(f, "</APPLET>\n");    
}

void ClassGraphWriter::writeText(FILE *f, Entry *e)
{
    int i;
    Entry *c;

    for(i = 0; i < max(1, e->baseclasses.size()); i++)
	{
	if(i < e->baseclasses.size() && e->baseclasses[i]->section == INTERFACE_SEC)
	    {
	    if(i == 0)
		fprintf(f, "<H3>%s</H3>\n", e->hname);
	    continue;
	    }
	fprintf(f, "<H3>%s\n", e->hname);
	c = e;
	while(c->baseclasses.size() > 0)
	    {
	    if(c == e)
		c = c->baseclasses[i];
	    else
		c = c->baseclasses[0];
	    if(c)
		fprintf(f, "- <A HREF=\"%s\">%s</A>\n",
		    c->fileName.c_str(),
		    c->hname);
	    else
		fprintf(f, "- %s</A>\n", c->name.c_str());  
	    }
	fprintf(f, "</H3>\n");
	}
}

void ClassGraphWriter::write(FILE *f, Entry *e)
{
    if(javaGraphs)
	writeJava(f, e);
    else 
	writeText(f, e);
}

void ClassGraphWriter::writeImplements(FILE *f, Entry *e)
{
    int i, first = 1;
    Entry *c;

    for(i = 0; i < e->baseclasses.size(); i++)
	{
	if(e->baseclasses[i]->section != INTERFACE_SEC)
	    continue;

	if(first)
	    fprintf(f, "<HR>\n<H2>Implements:</H2>\n");
	first = 0;

	c = e->baseclasses[i];
	if(i > 0)
	    fprintf(f, ", ");
	if(c)
	    fprintf(f, "<A HREF=\"%s\">%s</A>", c->fileName.c_str(),
		(const char*)(c->hname));

	}
}

void copyright(FILE *f)
{
    fprintf(f, "<HR>\n");
    if(footer.length() <= 0 && ownFooter == 0)
	{
	fprintf(f, "<BR>\n");
	fprintf(f, "This page was generated with the help of <A HREF=\"http://www.linuxsupportline.com/~doc++\">DOC++</A>.\n");
	fprintf(f, "</BODY>\n");
	}
    else
	fprintf(f, "%s", footer.c_str());
}

extern char *strToHtml(McString &in, char *dest = 0, Entry* ct = 0, int withLinks = 0, int pureCode = 0);
extern char *seeToHtml(McString &in, Entry* ct = 0);

void entry2link(McString& u, Entry *ref, const char *linkname)
{
    Entry *globref = ref;

    if(globref != root)
	while(globref->parent && globref->parent != root &&
	     !globref->fileName.length())
	    globref = globref->parent;
    if(ref->fileName.length())
	{
	u += "<!1><A HREF=\"";
	u += ref->fileName;
	if(!ref->ownPage)
	    {
	    u += "#DOC.";
	    ref->dumpNumber(u);
	    }
	u += "\">";
	if(linkname)
	    u += linkname;
	else
	    u += ref->hname;
	u += "</A>";
	}
    else
	if(globref)
	    {
	    u += "<!2><A HREF=\"" + globref->fileName + "#DOC.";
	    ref->dumpNumber(u);
	    u += "\">";
	    if(linkname)
		u += linkname;
	    else
		u += ref->hname;
	    u += "</A>";
	    }
}  

void writeTOCentry(McString& out, Entry *e, int memo, int dup = 0)
{ 
    McString link = "";

    if(e->fileName.length())
	{
	if(dup && e->parent)
	    link += "in ";
	link += "<A HREF=\"";
	if(e->parent)
	    if(e->ownPage)
		{
		if(e->fileName.length() > 0)
	    	    link += e->fileName;
                else
	    	    {
		    link += e->hname;
		    link += htmlSuffix;
		    }
		}
	    else
		{
		if(e->parent->fileName.length() > 0)
		    link += e->parent->fileName;
		else
		    {
		    link += e->parent->hname;
		    link += htmlSuffix;
		    }
		link += "#";
		link += e->hname;
		}
	else
	    {
	    link += GENERAL_NAME;
	    link += htmlSuffix;
	    link += "#";
	    link += e->hname;
	    }
	link += "\">";

	if(dup && e->parent)
	    {
	    if(e->parent->hname && strlen(e->parent->hname) > 0)
		link += (char *)(e->parent->hname);
	    else
		{
		link += "duplicate";
#ifdef DEBUG
		if(verb)
		    {
		    printf("Duff link from: \n");
		    e->dump(stdout);
		    }
#endif
		}
	    }
	else
	    if(java && e->section == PACKAGE_SEC && e->name.length())
		link += e->hname;
	    else
		link += e->fullName;

	link += "</A>";
	}
    else
	entry2link(link, e, (const char *)(e->hname));
    out += link;
    if(memo && !java)
	if(e->memo.length())
	    {
	    out += " <I>";
	    out += (const char *)e->hmemo;
	    out += "</I>\n";
	    }
         else
	    out += '\n';    
}

void writeTOCentry(FILE *f, Entry *e, int memo, int dup)
{ 
    McString out;

    writeTOCentry(out, e, memo, dup);
    fprintf(f, "%s", out.c_str());
}

void writeHIERentry(FILE *f, Entry *k, int memo)
{
    int i;

    fprintf(f, "<LI>");
    writeTOCentry(f, k, memo);
    if(k->pubChilds.size() || k->proChilds.size())
	{
	fprintf(f, "<UL>\n");
	for(i = 0; i < k->pubChilds.size(); i++)
	{
	    writeHIERentry(f, k->pubChilds[i], memo);
	}
	for(i = 0; i < k->proChilds.size(); i++)
	{
	    writeHIERentry(f, k->proChilds[i], memo);
	}
	fprintf(f, "</UL>\n");
	}
}

struct {
    int sec;
    const char *name;
} toc_sections[] = {
    { MANUAL_SEC,	"General stuff" },
    { PACKAGE_SEC,	"Packages" },
    { NAMESPACE_SEC,	"Namespaces" },
    { CLASS_SEC,	"Classes" },
    { INTERFACE_SEC,	"Interfaces" },
    { FUNCTION_SEC,	"Functions" },
    { VARIABLE_SEC,	"Variables" },
    { MACRO_SEC,	"Macros" },
    { UNION_SEC,	"Enums, Unions, Structs" },
    { TYPEDEF_SEC,	"Typedefs" },
    { 0, 0 }
};

void writeTOCRec(TOClist& list, FILE *f, Entry *root, int section, int& first)
{ 
    int i, output = root->section == toc_sections[section].sec &&
	root->name.length();

    // Want to recurse down em all, but subtree only if output root
    TOClist *sub = 0;
    if(root->sublist.size())
	{
	if(output)
	    sub = new TOClist();

	for(i = 0; i < root->sublist.size(); i++)
	    {
	    if(output)
		writeTOCRec(*sub, f, root->sublist[i], section, first);
	    else
		writeTOCRec(list, f, root->sublist[i], section, first);
	    }
	}
    if(output)
	list.addEntry(root, sub);
}

void writeTOC(FILE *f)
{
    int first = 1;

    fprintf(f, "%s", indexHeader.c_str());
    fprintf(f, "\n<H1>Table of Contents</H1>\n");

    for(int k = 0; toc_sections[k].name ; k++)
	{
	first = 1;
	TOClist list;
	writeTOCRec(list, f, root, k, first);
	if(list.size() > 0)
	    {
	    list.sort();
	    fprintf(f, "<H2>%s</H2>\n", toc_sections[k].name);
	    fprintf(f, "<UL>\n");
	    list.write(f);
	    fprintf(f, "</UL>\n");
	    }
	// Javadoc compatibility: fairly sure this is what it does.
	if(java && toc_sections[k].sec == INTERFACE_SEC)
	    break;
	}
    fprintf(f, "%s", indexFooter.c_str());
    copyright(f);
}

void writeHIERrec(FILE *f, Entry *root)
{
    int i;

    for(i = 0; i < root->sublist.size(); i++)
	{
	if(root->sublist[i]->baseclasses.size() == 0)
	    if(root->sublist[i]->isClass() &&
		root->sublist[i]->proBaseclasses.size() == 0 &&
		root->sublist[i]->pubBaseclasses.size() == 0)
		writeHIERentry(f, root->sublist[i], 0);
	writeHIERrec(f, root->sublist[i]);
	}
}

void writeHIER(FILE *f)
{
    fprintf(f, "%s", hierHeader.c_str());

    fprintf(f, "<H1>Hierarchy of Classes</H1>\n");
    fprintf(f, "<UL>\n");
    writeHIERrec(f, root);
    fprintf(f, "</UL>\n");

    fprintf(f, "%s", hierFooter.c_str());

    copyright(f);
}

void writeHIERrecJava(FILE *f, Entry *root)
{
    int i;

    for(i = 0; i < root->sublist.size(); i++)
	{
	if(root->sublist[i]->baseclasses.size() == 0)
	    if(root->sublist[i]->isClass() &&
		root->sublist[i]->proBaseclasses.size() == 0 &&
		root->sublist[i]->pubBaseclasses.size() == 0)
		ClassGraphWriter::writeJava(f, root->sublist[i], 0);
	writeHIERrecJava(f, root->sublist[i]);
	}
}

void writeHIERjava(FILE *f)
{
    fprintf(f, "%s", hierHeader.c_str());
    fprintf(f, "<H1>Hierarchy of classes</H1>\n");
    fprintf(f, "<UL>\n");
    writeHIERrecJava(f, root);
    fprintf(f, "</UL>\n");
    fprintf(f, "<I><A HREF=\"index.%s\"> alphabetic index</A></I><P>", htmlSuffix);
    copyright(f);
}

/** This class keeps track of overloading relationships. Insert mebers
    using addMember. It returns NULL in case this is a new Member, != 0
    otherwise.
*/
class MemberList {
	McHashTable<char *, Entry *> list;
    public:
	MemberList() : list((Entry *)NULL)
	    {}
	/** Add a new member. Returns NULL, if no compatible member
	    has yet occurred.
	*/
	Entry *addMember(Entry *e, Entry *father)
	    {
	    McString signature = e->name;
	    char *tmp;

	    if(e->language == LANG_JAVA)
		signature += e->args;
	    tmp = strdup(signature.c_str());
	    Entry **val = list.insert(tmp);
	    if(*val == 0)
		{
		*val = father;
		return 0;
		}
	    else
		{
#ifdef DEBUG
		if(verb)
		    printf("Member `%s' was there\n", tmp);
#endif
		free(tmp);
		}
	    return *val;
	    }
};

void showSubMembers(FILE *f, Entry *entry);

void showMembers(Entry *e, FILE *f, int links, MemberList *ignore = 0)
{
    static struct {
	char *heading;
	int protection;
	int secMask;
    } sections[] = {
      { "<DT><H3>Public Classes</H3><DD>", PUBL, CLASS_SEC | UNION_SEC },
      { "<DT><H3>Public Fields</H3><DD>", PUBL, VARIABLE_SEC },
      { "<DT><H3>Public Methods</H3><DD>", PUBL, FUNCTION_SEC},
      { "<DT><H3>Public</H3><DD>", PUBL, ~(CLASS_SEC | VARIABLE_SEC | FUNCTION_SEC | UNION_SEC) },
      { "<DT><H3>Protected Classes</H3><DD>", PROT, CLASS_SEC | UNION_SEC },
      { "<DT><H3>Protected Fields</H3><DD>", PROT, VARIABLE_SEC },
      { "<DT><H3>Protected Methods</H3><DD>", PROT, FUNCTION_SEC },
      { "<DT><H3>Protected</H3><DD>", PROT, ~(CLASS_SEC | VARIABLE_SEC | FUNCTION_SEC | UNION_SEC) },
      { "<DT><H3>Private Classes</H3><DD>", PRIV, CLASS_SEC | UNION_SEC },
      { "<DT><H3>Private Fields</H3><DD>", PRIV, VARIABLE_SEC },
      { "<DT><H3>Private Methods</H3><DD>", PRIV, FUNCTION_SEC },
      { "<DT><H3>Private</H3><DD>", PRIV, ~(CLASS_SEC | VARIABLE_SEC | FUNCTION_SEC | UNION_SEC) },
      { 0, 0, 0 }};

    int i;
    Entry *tmp, *othersFather;
    MemberWriter *memberWriter;
    bool ignoreThisOne;

    fprintf(f, "<DL>\n");
    if(withTables && links)
	memberWriter = new MemberWriterTable();
    else
	memberWriter = new MemberWriter();
    for(i = 0; sections[i].heading; i++)
	if(withPrivate || sections[i].protection != PRIV)
	    {
	    memberWriter->startList(f, sections[i].heading, links);
	    for(tmp = e->sub; tmp; tmp = tmp->next)
		{
		if(tmp->protection == sections[i].protection &&
		    (tmp->section & sections[i].secMask))
		    if(links ||
			(tmp->name != e->name && (tmp->name[0] != '~' ||
			    strcmp(tmp->name.c_str() + 1, e->name.c_str()) != 0)))
			{
			ignoreThisOne = false;
			if(ignore)
			    {
			    othersFather = ignore->addMember(tmp, e);
			    if(othersFather != 0 && othersFather != e)
				ignoreThisOne = true;
			    }
			bool withSub = (tmp->section == MANUAL_SEC);
			if(!ignoreThisOne || links)
			    memberWriter->addMember(tmp, links, withSub);
			}
		}
	    memberWriter->endList();
	    }

    delete memberWriter;
    fprintf(f, "</DL>\n");
}

void writeInherited(Entry *k, FILE *f, MemberList *list = 0)
{
    int i;

    showMembers(k, f, 0, list);
    for(i = 0; i < k->baseclasses.size(); i++)
	{
	fprintf(f, "<HR><H3>Inherited from <A HREF=\"%s\">%s</A>:</H3>\n",
	    (k->baseclasses)[i]->fileName.c_str(),
	    (k->baseclasses)[i]->hname);
	writeInherited((k->baseclasses)[i], f, list);
	}
}

/** This function writes the @-fields (except @memo, @name) of the
    specified entry
*/
void writeTags(FILE *f, Entry *entry)
{
    int i, k;

    fprintf(f, "<DL><DT>");

    if(entry->exception.size())
	{
	fprintf(f, "<DT><B>Throws:</B><DD>");
	for(i = 0; i < entry->exception.size(); i++)
	    {
	    k = 0;
	    McString s;
	    while(k < entry->exception[i]->length() &&
		(myisalnum((*entry->exception[i])[k]) ||
		(*entry->exception[i])[k] == '_' ||
		(*entry->exception[i])[k] == '<' ||
		(*entry->exception[i])[k] == '>' ||
		(*entry->exception[i])[k] == ':'))
		s += (*entry->exception[i])[k++];
	    fprintf(f, "<B>%s</B> ", strToHtml(s, 0, entry, 1, 1));
	    while(k < entry->exception[i]->length())
		fprintf(f, "%c", (*entry->exception[i])[k++]);
	    fprintf(f, "<BR>");
	    }
	}

    if(entry->precondition.size())
	{
	fprintf(f, "<DT><B>Preconditions:</B><DD>");
	for(i = 0; i < entry->precondition.size(); i++)
	    {
	    k = 0;
	    fprintf(f, "  ");
	    while(k < entry->precondition[i]->length())
		fprintf(f, "%c", (*entry->precondition[i])[k++]);
	    fprintf(f, "<BR>");
	    }
	}

    if(entry->postcondition.size())
	{
	fprintf(f, "<DT><B>Postconditions:</B><DD>");
	for(i = 0; i < entry->postcondition.size(); i++)
	    {
	    k = 0;
	    fprintf(f, "  ");
	    while(k < entry->postcondition[i]->length())
		fprintf(f, "%c", (*entry->postcondition[i])[k++]);
	    fprintf(f, "<BR>");
	    }
	}

    if(entry->invariant.size())
	{
	fprintf(f, "<DT><B>Invariants:</B><DD>");
	for(i = 0; i < entry->invariant.size(); i++)
	    {
	    k = 0;
	    fprintf(f, "  ");
	    while(k < entry->invariant[i]->length())
		fprintf(f, "%c", (*entry->invariant[i])[k++]);
	    fprintf(f, "<BR>");
	    }
	}

    if(entry->retrn.size())
	{
	fprintf(f, "<DT><B>Returns:</B><DD>");
	for(i = 0; i < entry->retrn.size(); i++)
	    {
	    k = 0;
	    fprintf(f, "  ");
	    while(k < entry->retrn[i]->length())
		fprintf(f, "%c", (*entry->retrn[i])[k++]);
	    fprintf(f, "<BR>");
	    }
	}

    if(entry->param.size())
	{
	fprintf(f, "<DT><B>Parameters:</B><DD>");
	for(k = 0; k < entry->param.size(); k++)
	    {
	    i = 0;
	    fprintf(f,"<B>");
	    while(i < entry->param[k]->length() &&
		(myisalnum((*entry->param[k])[i]) || 
		(*entry->param[k])[i] == '_'))
		fprintf(f, "%c", (*entry->param[k])[i++]);
	    fprintf(f, "</B> - ");

	    while(i < entry->param[k]->length())
		fprintf(f, "%c", (*entry->param[k])[i++]);
	    fprintf(f, "<BR>");
	    }
	}

    if(entry->author.length())
	fprintf(f, "<DT><B>Author:</B><DD>%s\n", entry->author.c_str());

    if(entry->version.length())
	fprintf(f, "<DT><B>Version:</B><DD>%s\n", entry->version.c_str());

    if(entry->see.size())
	{
	fprintf(f, "<DT><B>See Also:</B><DD>");
	for(k = 0; k < entry->see.size(); k++)
	    if(entry->see[k]->length())
		fprintf(f, "%s<BR>", seeToHtml(*entry->see[k], entry));
	}

    fprintf(f, "<DD></DL><P>");
}

void writeDoc(FILE *f, Entry *entry)
{  
    if(entry->ownPage)
	return;
    int toBeDocumented = subEntryIsToBeDocumented(entry);
    Entry *tmp;

    if(toBeDocumented)
	{
	char *args = (entry->hargs);
	char *type = (entry->htype);
	printRefLabel(f, entry);
	fprintf(f, "<DT>");
	fprintf(f, "<IMG ALT=\"o\" BORDER=0 SRC=icon2.gif><TT><B>%s %s",
	    type, (char *)(entry->hname));
	fprintf(f, "%s", args);	
	fprintf(f, "</B></TT>\n");

	if(entry->doc.length() > 0)
	    fprintf(f, "<DD>%s\n", (char *)(entry->hdoc));
	else
	    if(entry->memo.length() > 0)
		fprintf(f, "<DD>%s\n", (char *)(entry->hmemo));

	writeTags(f, entry);
	}

    if(entry->sub)
	{
	fprintf(f, "<DL>\n");
	for(tmp = entry->sub; tmp; tmp = tmp->next)
	    writeDoc(f, tmp);
	fprintf(f, "</DL>\n");
	}
}

void writeHeader(Entry *e, FILE *f)
{
    McString tmp;

    if(ownHeader == 0)
	{
	fprintf(f, "<HTML>\n");
	fprintf(f, "<HEAD>\n");
	fprintf(f, "   <TITLE>%s %s</TITLE>\n", e->type.c_str(), e->fullName.c_str());
	fprintf(f, "   <META NAME=\"GENERATOR\" CONTENT=\"DOC++ " DOCXX_VERSION "\">\n");
	fprintf(f, "</HEAD>\n");
	fprintf(f, "%s\n", pageHeader.c_str());
	}
    else
	fprintf(f, "%s\n", header.c_str());

    if(showFilenames && e->section != PACKAGE_SEC && e->section != MANUAL_SEC)
	fprintf(f, "In file %s:", e->file.c_str());

    if(e->section == PACKAGE_SEC)
	{
	e->getPackage(tmp);
	fprintf(f, "<H2>Package %s</H2>", tmp.c_str());
	}
    else
	{
	fprintf(f, "<H2>");
	if(e->language == LANG_JAVA)
	    {
	    e->getPackage(tmp);
	    fprintf(f, "%s <A HREF=\"#DOC.DOCU\">%s.%s", e->htype,
		tmp.c_str(), e->hname);
	    }
	else
	    fprintf(f, "%s <A HREF=\"#DOC.DOCU\">%s", e->htype,
		(e->fullName.length() ? e->fullName.c_str() : e->hname));

	fprintf(f, "</A>%s</H2>\n", (const char *)(e->hargs));
	}
    htmlComment(f, e, MEMO);
}

void writeManPage(Entry *e, FILE *f)
{
    int i, numChilds, numParents = 0;
    Entry *c, *tmp;
    MemberList list;
    MemberWriter memberWriter;

#ifdef DEBUG
    if(verb)
	printf("Writing page for `%s %s'\n", e->type.c_str(),
	    e->name.c_str());
#endif

    // the page header
    writeHeader(e, f);

    if(e->isClass())            // is it really a class?
	{ 
	// the inheritance
	numChilds = e->pubChilds.size() + e->proChilds.size() + e->priChilds.size();
	if(e->language == LANG_JAVA)
	    numParents = e->pubBaseclasses.size()+ e->proBaseclasses.size() +
		e->otherPubBaseclasses.size() + e->otherProBaseclasses.size();
	else
	    numParents = e->pubBaseclasses.size() + e->priBaseclasses.size() +
		e->proBaseclasses.size() + e->otherPubBaseclasses.size() +
		e->otherPriBaseclasses.size() + e->otherProBaseclasses.size();

	if(numParents > 0 || numChilds > 0 || trivialGraphs)
	    {
	    fprintf(f, "<HR>\n\n<H2>Inheritance:</H2>\n");
	    ClassGraphWriter::write(f, e);
	    if(e->language == LANG_JAVA && e->implements.size() > 0)
		ClassGraphWriter::writeImplements(f, e);
	    }

	// the members 
	if(e->sub)
	    {
	    fprintf(f, "<HR>\n\n");
	    showMembers(e, f, 1, &list);
	    }
	}
    else                       // this is not a class
	{ 
	if(e->sub)
	    {
	    fprintf(f, "\n<HR>\n");
	    memberWriter.startList(f, " ", 1);
	    for(tmp = e->sub; tmp; tmp = tmp->next)
		memberWriter.addMember(tmp, 1, 0);
	    }
	memberWriter.endList();
	}

    // the inherited members
    if(e->isClass())            // is it really a class?
	if(showInherited)
	    for(i = 0; i < e->baseclasses.size(); i++)
		{
		fprintf(f, "<HR><H3>Inherited from <A HREF=\"%s\">%s</A>:</H3>\n",
		    (e->baseclasses)[i]->fileName.c_str(),
		    (e->baseclasses)[i]->hname);
		writeInherited((e->baseclasses)[i], f, &list);
		}

    // the documentation
    fprintf(f, "\n<A NAME=\"DOC.DOCU\">\n");
    fprintf(f, "<HR>\n<H2>Documentation</H2>\n");
    htmlComment (f, e, (DOC | MEMO));
    fprintf(f, "<DL>\n");
    for(tmp = e->sub; tmp; tmp = tmp->next)
	writeDoc(f, tmp);
    fprintf(f, "</DL>\n");

    // the childrens
    if(e->isClass())            // is it really a class?
	if(e->pubChilds.size() || e->proChilds.size())
	    {
	    fprintf(f, "<HR>\n<DL><DT><B>Direct child classes:\n</B><DD>");
	    c = e;
	    for(i = 0; i < c->pubChilds.size(); i++)
		fprintf(f, "<A HREF=\"%s\">%s</A><BR>\n",
		    c->pubChilds[i]->fileName.c_str(),
		    c->pubChilds[i]->hname);
	    for(i = 0; i < c->proChilds.size(); i++)
		fprintf(f, "<A HREF=\"%s\">%s</A><BR>\n",
		    c->proChilds[i]->fileName.c_str(),
		    c->proChilds[i]->hname);
	    fprintf(f, "</DL>\n\n");
	    }
	else
	    fprintf(f, "\n<HR><DL><DT><B>This class has no child classes.</B></DL>\n\n");

    writeTags(f, e);
    if(javaGraphs)
	fprintf(f, "%s", pageFooterJava.c_str());
    else
	fprintf(f, "%s", pageFooter.c_str());

    copyright(f);
}

static FILE *myOpen(const char *dir, const McString& name)
{
    FILE *f;
    McString buf = dir;
    buf += PATH_DELIMITER;
    buf += name;
#ifdef DEBUG
    if(verb)
	printf("Opening `%s' to write\n", (char *)buf);
#endif
    if(!(f = fopen((char *)buf, "wb")))
	{
	fprintf(stderr, "Cannot open `%s'\n", (char *)buf);
	exit(-1);
	}
    return f;
}

void makeHtmlNames(Entry *entry)
{
    Entry *tmp;
    entry->hname = strToHtml(entry->name, 0, entry, 0);
    if(entry->section == EMPTY_SEC)
	entry->section = PACKAGE_SEC;

    if(entry->sub)
	for(tmp = entry->sub; tmp; tmp = tmp->next)
	    makeHtmlNames(tmp);
}

// This creates names for all those with their own pages
void makeFileNames(Entry *entry)
{
    Entry *tmp;
    McString file;

    if(entry->ownPage)
	{
	if(entry->language == LANG_JAVA)
	    {
	    McString pack;
	    if(entry->section == PACKAGE_SEC)
		file = "package-";
	    entry->getPackage(pack);
	    if(pack.length())
		{
		file += pack;
		if(entry->section != PACKAGE_SEC)
		    file += '.'; 
		}
	    }
	if(entry->language != LANG_JAVA || entry->section != PACKAGE_SEC)
	    {
	    if(entry->name.length())
		file += entry->name;
	    else
		if(entry->section == UNION_SEC) // Struct, union, enum
		    {
		    file += entry->type;
#ifdef DEBUG
		    if(verb)
			{	// This is for unnamed enums / unions etc.
			printf("Making name from type for:\n");
			entry->dump(stdout);
			}
#endif
		    }
	    }
	entry->fileName = makeFileName(file, entry);
	}
    if(entry->sub)
	for(tmp = entry->sub; tmp; tmp = tmp->next)
	    makeFileNames(tmp);
}

// This creates filenames for items from their parents
void inheritFileNames(Entry *entry)
{
    Entry *tmp;

    if(!entry->ownPage)
	{
	tmp = entry;
	while(tmp->parent && tmp->fileName.length() == 0)
	    tmp = tmp->parent;
	if(tmp != entry)
	    {
	    entry->fileName = (char *)tmp->fileName;
#ifdef DEBUG
	    if(verb)
		printf("Copied `%s %s' into `%s'\n", (char *)entry->type,
		    (char *)entry->name, (char *)entry->fileName);
#endif
	    }
	}

    if(entry->sub)
	for(tmp = entry->sub; tmp; tmp = tmp->next)
	    inheritFileNames(tmp);
}

// This marks those names that can't be attached to a class
// It also junks redundant floating comments as a bad job...
void relocateNames(Entry *root)
{
    Entry *e, *next;
    e = root->sub;
    while(e)
	{
	next = e->next;
	if(!e->ownPage)
	    {
	    if(useGeneral)
		{
#ifdef DEBUG
		if(verb)
		    {
		    printf("`%s' hanging from root with no page\n",
			(char *)e->name);
		    e->dump(stdout);
		    }
#endif
		e->general = true;
		}
	    else
		{
		root->removeSub(e);
		if(!fastNotSmall)
		    delete(e);
		}
	    }
	e = next;
	}	
}

// Decide what entries should have their's own page
void decideAboutOwnPages(Entry *entry)
{
    Entry *tmp;

    if(entry->section == EMPTY_SEC)
	entry->section = MANUAL_SEC;

    // The Quantel extension bit is to stop annoying stuff like the RCS info
    // (often in a FED fold) from turning up in the data...
    if(entry->sub || (entry->doc.length() > 1 && !QuantelExtn) || entry->isClass())
	entry->ownPage = true;

    if(entry->sub && (entry->section == MANUAL_SEC ||
	entry->section == PACKAGE_SEC ||
	entry->section == NAMESPACE_SEC || (entry->section & CLASS_SEC)))
	for(tmp = entry->sub; tmp; tmp = tmp->next)
	    if(!(entry->section & CLASS_SEC) || ((entry->section & CLASS_SEC) &&
		(tmp->section & CLASS_SEC)))
		decideAboutOwnPages(tmp);
}

void makeHtml(Entry *entry)
{
    Entry *tmp;
    int i;

    entry->hmemo = strToHtml(entry->memo, 0, entry, 0);
    entry->hdoc = strToHtml(entry->doc, 0, entry, 0);
    entry->hargs = strToHtml(entry->args, 0, entry, 1, 1);
    entry->htype = strToHtml(entry->type, 0, entry, 1, 1);
    entry->author = strToHtml(entry->author, 0, entry, 0);
    entry->version = strToHtml(entry->version, 0, entry, 0);

    for(i = 0; i < entry->retrn.size(); i++)
	*entry->retrn[i] = strToHtml(*entry->retrn[i], 0, entry, 0);

    for(i = 0; i < entry->param.size(); i++)
	*entry->param[i] = strToHtml(*entry->param[i], 0, entry, 0);

    for(i = 0; i < entry->exception.size(); i++)
	*entry->exception[i] = strToHtml(*entry->exception[i], 0, entry, 0);

    if(entry->sub)
	for(tmp = entry->sub; tmp; tmp = tmp->next)
	    makeHtml(tmp);
}

void writePageSub(FILE *f, Entry *e)
{
    McString htype, hargs;

#ifdef DEBUG
    if(verb)
	printf("Writing sub for `%s %s' to file `%s'\n", (char *)e->type,
	    (char *)e->name, (char *)e->fileName);
#endif

    if(withTables)
	fprintf(f, "<TR><TD VALIGN=top>");
    else
	fprintf(f, "<DT>\n");

    // Show the signature if this is a leaf node (meaning its full
    // documentation will be on the current page)
    if(e->fileName.length() == 0 || e->general)
	{
	htype = e->htype;
	hargs = e->hargs;
	}

    // Show the bullet symbol.
    fprintf(f, "\n<IMG ALT=\"o\" BORDER=0 SRC=icon1.gif>");

    if(htype.length() > 0)
	fprintf(f, "%s", htype.c_str());

    // label the symbol being here
    fprintf(f, "<A NAME=\"%s\"></A>\n", e->name.c_str());

    // create a cross reference to the symbol if elsewhere.
    if(!e->general && e->fileName.length() > 0)
	fprintf(f, "<A HREF=%s><B>%s</B></A>", e->fileName.c_str(),
	    e->hname);
    else
	fprintf(f, "<B>%s</B>", e->hname);

    if(hargs.length() > 0)
	fprintf(f, "%s", hargs.c_str());

    if(withTables)
	{
	fprintf(f, "</TD><TD><BR>\n");
	if(e->hmemo)
	    fprintf(f, "<I>%s</I>\n", (const char *)e->hmemo);
	fprintf(f, "</TD></TR>");
	}
    else
	if(e->hmemo)
	    fprintf(f, "<DD><I>%s</I>\n", (const char *)e->hmemo);
}

void writeManPageRec(const char *dir, Entry *e)
{
    Entry *tmp;
    FILE *f;

#ifdef DEBUG
    if(verb)
	printf("Writing `%s %s' to file `%s'\n", (char *)e->type,
	    (char *)e->name, (char *)e->fileName);
#endif

    if(e->general)
	writePageSub(generalf, e);

    if(e->ownPage)
	{
#ifdef DEBUG
	if((e->fileName == htmlSuffix ||
	    e->fileName == (const char *)".2.html") && verb)
	    {
	    fprintf(stderr, "Warning: weird filename `%s' for `%s %s'\n",
		(char *)e->fileName, (char *)e->type, (char *)e->name);
	    e->dump(stdout);
	    }
#endif
	if(!(f = myOpen(dir, e->fileName)))
	    {
	    fprintf(stderr, "Cannot open `%s' for writing\n", (char *)e->fileName);
	    return;
	    }
	if(e->section != MANUAL_SEC && e->section != PACKAGE_SEC)
	    writeManPage(e, f);
	else
	    {
	    writeHeader(e, f);
	    if(e->sub)
		{
		if(withTables)
		    fprintf(f, "\n<TABLE>\n");
		else
		    fprintf(f, "\n<HR>\n<DL>\n");
		for(tmp = e->sub; tmp; tmp = tmp->next)
		    writePageSub(f, tmp);
		if(withTables)
		    fprintf(f, "\n</TABLE>\n");
		else
		    fprintf(f, "</DL>\n");
		}

	    fprintf(f, "<A NAME=\"DOC.DOCU\">\n");
	    htmlComment(f, e, (DOC | MEMO));
	    writeTags(f, e);
	    fprintf(f, "%s", pageFooter.c_str());
	    copyright(f);
	    }
	fclose(f);
	}
    if(e->sub && (e->section == MANUAL_SEC || e->section == PACKAGE_SEC ||
	e->section == NAMESPACE_SEC || (e->section & CLASS_SEC)))
	for(tmp = e->sub; tmp; tmp = tmp->next)
	    if(!(e->section & CLASS_SEC) || ((e->section & CLASS_SEC) &&
		(tmp->section & CLASS_SEC)))
		writeManPageRec(dir, tmp);
}

static void dumpFile(const char *dir, const char *name, const unsigned char *data, int size)
{
    FILE *f = myOpen(dir, name);
    fwrite(data, 1, size, f);
    fclose(f);
}

void readTemplates()
{
    FILE *f;
    int c;

    if((f = fopen("indexHeader.inc", "r")))
	{
	indexHeader = " ";
	while((c = fgetc(f)) != EOF)
	    indexHeader += c;
	fclose(f);
	}
    if((f = fopen("indexFooter.inc", "r")))
	{
	indexFooter = " ";
	while((c = fgetc(f)) != EOF)
	    indexFooter += c;
	fclose(f);
	}
    if((f = fopen("hierHeader.inc", "r")))
	{
	hierHeader = " ";
	while((c = fgetc(f)) != EOF)
	    hierHeader += c;
	fclose(f);
	}
    if((f = fopen("hierFooter.inc", "r")))
	{
	hierFooter = " ";
	while((c = fgetc(f)) != EOF)
	    hierFooter += c;
	fclose(f);
	}
    if((f = fopen("classHeader.inc", "r")))
	{
	pageHeader = " ";
	while((c = fgetc(f)) != EOF)
	    pageHeader += c;
	fclose(f);
	}
    if((f = fopen("classFooter.inc", "r")))
	{
	pageFooter = " ";
	while((c = fgetc(f)) != EOF)
	    pageFooter += c;
	fclose(f);
	}
}

void doHTML(char *dir, Entry *)
{
    FILE *f;
    Entry *tmp;
    int c;

    if(makedir(dir, 0755) != 0)
	if(errno == EEXIST)
	    {
	    FILE *exist = fopen(dir, "a");
	    if(exist)
		{
		fprintf(stderr, "`%s' file already exists\n", dir);
		fclose(exist);
		}
	    }
	else
	    fprintf(stderr, "Could not create `%s' directory\n", dir);

    decideAboutOwnPages(root);
    makeFileNames(root);

    // This is as items under root with no page inherit this name
    root->fileName = GENERAL_NAME;
    root->fileName += htmlSuffix;
    inheritFileNames(root);
    relocateNames(root);

    if(root->name.length() && root->section == MANUAL_SEC)
	{
	root->fileName = "index";
	root->fileName += htmlSuffix;
	}
    else
	if(root->sublist.size() == 1 &&		// Unconvinced about this...
	    root->sublist[0]->section == MANUAL_SEC)
	    {
	    root->sublist[0]->fileName = "index";
	    root->sublist[0]->fileName += htmlSuffix;
	    }

#ifdef DEBUG
    if(verb)
	{
	root->dump(stdout);
	for(tmp = root->sub; tmp; tmp = tmp->next)
	    tmp->dump(stdout);
	}
#endif

    if(verb)
	printf("Converting DOC++ to HTML...\n");

    if(root->section == EMPTY_SEC)
	root->section = MANUAL_SEC;
    for(tmp = root; tmp; tmp = tmp->next)
	makeHtmlNames(tmp);
    for(tmp = root; tmp; tmp = tmp->next)	// This takes the bulk of the time
	makeHtml(tmp);

    readTemplates();
    if(verb)
	printf("Writing files...\n");
    dumpFile(dir, "logo.gif", logo, sizeof(logo));
    dumpFile(dir, "icon1.gif", blueBall, sizeof(blueBall));
    dumpFile(dir, "icon2.gif", greyBall, sizeof(greyBall));
    dumpFile(dir, "down.gif", down, sizeof(down));

    if(ownFooter)
	{
	FILE *in = fopen(ownFooter, "r");
	if(!in && strcmp("none", ownFooter) && strlen(ownFooter) > 1)
	    fprintf(stderr, "Warning: Can't open `%s', producing no footer.\n",
		ownFooter);
	else
	    {
	    while((c = fgetc(in)) != EOF)
		footer += (char)c;
	    fclose(in);
	    }
	}

    if(ownHeader)
	{
	FILE *in = fopen(ownHeader, "r");
	if(!in && strcmp("none", ownHeader) && strlen(ownHeader) > 1)
	    fprintf(stderr, "Warning: Can't open `%s', producing no header.\n",
		ownHeader);
	else
	    {
	    while((c = fgetc(in)) != EOF)
		header += (char)c;
	    fclose(in);
	    }
	}

    if(javaGraphs)
	{
	dumpFile(dir, "ClassGraph.class", ClassGraph_class,
	    sizeof(ClassGraph_class));
	dumpFile(dir, "ClassGraphPanel.class", ClassGraphPanel_class,
	    sizeof(ClassGraphPanel_class));
	dumpFile(dir, "ClassLayout.class", ClassLayout_class,
	    sizeof(ClassLayout_class));
	dumpFile(dir, "NavigatorButton.class", NavigatorButton_class,
	    sizeof(NavigatorButton_class));
	}

    McString buf;

    // Table of contents
    buf = "index";
    buf += htmlSuffix;
    f = myOpen(dir, buf);
    writeTOC(f);
    fclose(f);      

    if(verb)
	printf("Writing TOC...\n");

    // Class heirarchy
    buf = "HIER";
    buf += htmlSuffix;
    f = myOpen(dir, buf);
    writeHIER(f);
    fclose(f);  

    // Java class heirarchy
    if(javaGraphs)
	{
	buf = "HIERjava";
	buf += htmlSuffix;
	f = myOpen(dir, buf);
	writeHIERjava(f);
	fclose(f);  
	}

    if(verb)
	printf("Writing Class Hierarchy...\n");

    /* Render all the rest of the pages to HTML, and output a 'General' file
       of floating #defines, global variables / functions etc.
    */

    // We don't want root page coming out as .html !
    root->ownPage = false;
    root->general = false;

    buf = GENERAL_NAME;
    buf+= htmlSuffix;
    generalf = myOpen(dir, buf);
    fprintf(generalf, "%s", generalHeader.c_str());
    if(withTables)
	fprintf(generalf, "\n<TABLE>\n");
    else
	fprintf(generalf, "\n<DL>\n");

    // Recursively write all pages
    writeManPageRec(dir, root);

    if(withTables)
	fprintf(generalf, "\n</TABLE>\n");
    else
	fprintf(generalf, "</DL>\n");

    fprintf(generalf, "%s", pageFooter.c_str());
    copyright(generalf);

    fclose(generalf);
}      
      
static void dumpEntry(FILE *out, Entry *entry, int)
{
    Entry *tmp;

    if(entry->sub)
	for(tmp = entry->sub; tmp; tmp = tmp->next)
    	    dumpEntry(out, tmp, 0);
}

void usermanHTML(char *, Entry *root)
{
    if(root->name.length())
	dumpEntry(out, root, 1);
    else
	for(root = root->sub; root; root = root->next)
	    dumpEntry(out, root, 1);
}
